""" Helpers are util functions which use the context """
from mrq.context import connections
import time


def ratelimit(key, limit, per=1, redis=None):
  """ Returns an integer with the number of available actions for the
  current period. If zero, rate was already reached. """

  if redis is None:
    redis = connections.redis

  # http://redis.io/commands/INCR
  now = int(time.time())

  k = "ratelimit:%s:%s" % (key, now // per)

  with redis.pipeline(transaction=True) as pipeline:
    pipeline.incr(k, 1)
    pipeline.expire(k, 10)
    value = pipeline.execute()

  current = int(value[0])

  if current >= limit:
    return 0
  else:
    return limit - current



#!/usr/bin/env python
import os

# Needed to make getaddrinfo() work in pymongo on Mac OS X
# Docs mention it's a better choice for Linux as well.
# This must be done asap in the worker
if "GEVENT_RESOLVER" not in os.environ:
  os.environ["GEVENT_RESOLVER"] = "ares"

# We must still monkey-patch the methods for job sub-pools.
from gevent import monkey
monkey.patch_all()

import sys
import argparse
import ujson as json
import datetime

sys.path.insert(0, os.getcwd())

from mrq import config, queue, utils
from mrq.context import set_current_config, set_current_job
from mrq.utils import load_class_by_path


def main():

  parser = argparse.ArgumentParser(description='Runs a task')

  cfg = config.get_config(parser=parser, config_type="run")
  cfg["is_cli"] = True
  set_current_config(cfg)

  if len(cfg["taskargs"]) == 1:
    params = json.loads(cfg["taskargs"][0])
  else:
    params = {}

    # mrq-run taskpath a 1 b 2 => {"a": "1", "b": "2"}
    for group in utils.group_iter(cfg["taskargs"], n=2):
      if len(group) != 2:
        print "Number of arguments wasn't even"
        sys.exit(1)
      params[group[0]] = group[1]

  if cfg["async"]:
    ret = queue.send_task(cfg["taskpath"], params, sync=False, queue=cfg["queue"])
    print ret
  else:
    worker_class = load_class_by_path(cfg["worker_class"])
    job = worker_class.job_class(None)
    job.data = {
      "path": cfg["taskpath"],
      "params": params,
      "queue": cfg["queue"]
    }
    job.datestarted = datetime.datetime.utcnow()
    set_current_job(job)
    ret = job.perform()
    print json.dumps(ret)

if __name__ == "__main__":
  main()



from .utils import load_class_by_path, group_iter, memoize
from .context import connections, get_current_config, log
import time


@memoize
def _redis_command_zaddbyscore():
  """ Increments multiple keys in a sorted set & returns them """

  return connections.redis.register_script("""
local zset = KEYS[1]
local min = ARGV[1]
local max = ARGV[2]
local offset = ARGV[3]
local count = ARGV[4]
local score = ARGV[5]

local data = redis.call('zrangebyscore', zset, min, max, 'LIMIT', offset, count)
for i, member in pairs(data) do
  redis.call('zadd', zset, score, member)
end

return data
  """)


@memoize
def _redis_command_zpopbyscore():
  """ Pops multiple keys by score """

  return connections.redis.register_script("""
local zset = KEYS[1]
local min = ARGV[1]
local max = ARGV[2]
local offset = ARGV[3]
local count = ARGV[4]

local data = redis.call('zrangebyscore', zset, min, max, 'LIMIT', offset, count)
if #data > 0 then
  redis.call('zremrangebyrank', zset, 0, #data - 1)
end

return data
  """)


@memoize
def _redis_command_lpopsafe():
  """ Increments multiple keys in a sorted set & returns them """

  return connections.redis.register_script("""
local key = KEYS[1]
local zset_started = KEYS[2]
local count = ARGV[1]
local now = ARGV[2]
local left = ARGV[3]
local data = {}
local current = nil

for i=1, count do
  if left == '1' then
    current = redis.call('lpop', key)
  else
    current = redis.call('rpop', key)
  end
  if current == false then
    return data
  end
  data[i] = current
  redis.call('zadd', zset_started, now, current)
end

return data
""")


class Queue(object):

  is_raw = False
  is_timed = False
  is_sorted = False
  is_set = False
  is_reverse = False

  def __init__(self, queue_id):
    if isinstance(queue_id, Queue):
      self.id = queue_id.id  # TODO use __new__?
      self.is_reverse = queue_id.is_reverse
    else:
      if queue_id[-8:] == "_reverse":
        self.is_reverse = True
        queue_id = queue_id[:-8]
      self.id = queue_id

    if "_raw" in self.id:
      self.is_raw = True

    if "_set" in self.id:
      self.is_set = True
      self.is_raw = True

    if "_timed" in self.id:
      self.is_timed = True
      self.is_sorted = True

    if "_sorted" in self.id:
      self.is_sorted = True

  @property
  def redis_key(self):
    return "%s:q:%s" % (get_current_config()["redis_prefix"], self.id)

  @classmethod
  def redis_key_started(self):
    return "%s:s:started" % get_current_config()["redis_prefix"]

  def enqueue_job_ids(self, job_ids):

    if len(job_ids) == 0:
      return

    if self.is_raw:

      retry_queue = get_current_config().get("raw_queues", {}).get(self.id, {}).get("retry_queue", "default")

      queue = Queue(retry_queue)
    else:
      queue = self

    # ZSET
    if queue.is_sorted:

      if type(job_ids) is not dict and queue.is_timed:
        now = time.time()
        job_ids = {str(x): now for x in job_ids}

      connections.redis.zadd(queue.redis_key, **job_ids)

    # LIST
    else:
      connections.redis.rpush(queue.redis_key, *job_ids)

  def enqueue_raw_jobs(self, params_list):

    if not self.is_raw:
      raise Exception("Can't queue raw jobs in a regular queue")

    if len(params_list) == 0:
      return

    # ZSET
    if self.is_sorted:

      if type(params_list) is not dict and self.is_timed:
        now = time.time()
        params_list = {str(x): now for x in params_list}

      connections.redis.zadd(self.redis_key, **params_list)
    # SET
    elif self.is_set:
      connections.redis.sadd(self.redis_key, *params_list)
    # LIST
    else:
      connections.redis.rpush(self.redis_key, *params_list)

  def remove_raw_jobs(self, params_list):

    if not self.is_raw:
      raise Exception("Can't remove raw jobs in a regular queue")

    if len(params_list) == 0:
      return

    # ZSET
    if self.is_sorted:
      connections.redis.zrem(self.redis_key, *iter(params_list))

    # SET
    elif self.is_set:
      connections.redis.srem(self.redis_key, *params_list)

    else:
      # O(n)! Use with caution.
      for k in params_list:
        connections.redis.lrem(self.redis_key, 1, k)

  def size(self):

    # ZSET
    if self.is_sorted:
      return connections.redis.zcard(self.redis_key)
    # SET
    elif self.is_set:
      return connections.redis.scard(self.redis_key)
    # LIST
    else:
      return connections.redis.llen(self.redis_key)

  def count_jobs_to_dequeue(self):
    if self.is_timed:
      # timed ZSET
      return connections.redis.zcount(self.redis_key, "-inf", time.time())
    else:
      return self.size()

  def empty(self):
    return connections.redis.delete(self.redis_key)

  def list_job_ids(self, skip=0, limit=20):

    if self.is_raw:
      raise Exception("Can't list job ids from a raw queue")

    # ZSET
    if self.is_sorted:
      return connections.redis.zrange(self.redis_key, skip, skip + limit - 1)
    # SET
    elif self.is_set:
      return connections.redis.srandmember(self.redis_key, limit)
    # LIST
    else:
      return connections.redis.lrange(self.redis_key, skip, skip + limit - 1)

  def get_sorted_graph(self, start=0, stop=100, slices=100, include_inf=False, exact=False):

    if not self.is_sorted:
      raise Exception("Not a sorted queue")

    with connections.redis.pipeline(transaction=exact) as pipe:
      interval = float(stop - start) / slices
      for i in range(0, slices):
        pipe.zcount(self.redis_key, (start + i * interval), "(%s" % (start + (i + 1) * interval))
      if include_inf:
        pipe.zcount(self.redis_key, stop, "+inf")
        pipe.zcount(self.redis_key, "-inf", "(%s" % start)
      data = pipe.execute()

    if include_inf:
      return data[-1:] + data[:-1]

    return data

  @classmethod
  def all_active(cls):
    """ List active queues, based on their lengths in Redis. """

    prefix = get_current_config()["redis_prefix"]
    queues = []
    for key in connections.redis.keys():
      if key.startswith(prefix):
        queues.append(Queue(key[len(prefix) + 3:]))

    return queues

  @classmethod
  def all(cls):
    """ List all known queues (Raw + MongoDB)"""

    # Start with raw queues we know exist from the config
    queues = {x: 0 for x in get_current_config().get("raw_queues", {})}

    stats = list(connections.mongodb_jobs.mrq_jobs.aggregate([
      {"$match": {"status": "queued"}},
      {"$group": {"_id": "$queue", "jobs": {"$sum": 1}}}
    ])["result"])

    queues.update({x["_id"]: x["jobs"] for x in stats})

    return queues

  @classmethod
  def dequeue_jobs(self, queues, max_jobs=1, job_class=None, worker=None):
    """ Fetch a maximum of max_jobs from this worker's queues. """

    if job_class is None:
      from .job import Job
      job_class = Job

    queue_objects = [Queue(q) for q in queues]

    has_raw = any(q.is_raw or q.is_sorted for q in queue_objects)

    log.debug("Fetching %s jobs from Redis" % max_jobs)

    # When none of the queues is a raw queue, we can have an optimized mode where we BLPOP from the queues.
    if not has_raw:

      jobs = []

      if queue_objects[0].is_timed:

        if worker:
          worker.status = "spawn"

        queue = queue_objects[0]
        with connections.redis.pipeline(transaction=True) as pipe:
          pipe.zrange(queue.redis_key, 0, max_jobs)
          pipe.zremrangebyrank(queue.redis_key, 0, max_jobs)
          job_ids = pipe.execute()

        # From this point until job.fetch_and_start(), job is only local to this worker.
        # If we die here, job will be lost in redis without having been marked as "started".

        jobs = [job_class(_job_id, queue=queue.redis_key, start=True) for _job_id in job_ids]

      else:

        simulate_lost_jobs = get_current_config().get("simulate_lost_jobs")

        for queue in queue_objects:

          job_ids = _redis_command_lpopsafe()(keys=[
            queue.redis_key,
            Queue.redis_key_started()
          ], args=[
            max_jobs,
            time.time(),
            "0" if queue.is_reverse else "1"
          ])

          if len(job_ids) == 0:
            continue

          if simulate_lost_jobs:
            break

          if worker:
            worker.status = "spawn"

          jobs += [job_class(_job_id, queue=queue, start=True)
                   for _job_id in job_ids if _job_id]

          # Now the jobs have been marked as started in Mongo, we can remove them from the started queue.
          connections.redis.zrem(Queue.redis_key_started(), *job_ids)

          max_jobs -= len(job_ids)

          if max_jobs == 0:
            break

      return jobs

    else:

      if worker:
        worker.status = "spawn"

      jobs = []

      # Try to dequeue from each of the queues until we have filled max_jobs.
      for queue in queue_objects:

        queue_config = get_current_config().get("raw_queues", {}).get(queue.id, {})

        job_factory = queue_config.get("job_factory")
        if not job_factory and queue.is_raw:
          raise Exception("No job_factory configured for raw queue %s" % queue.id)

        params = []

        # ZSET with times
        if queue.is_timed:

          current_time = time.time()

          # When we have a pushback_seconds argument, we never pop items from the queue, instead
          # we push them back by an amount of time so that they don't get dequeued again until
          # the task finishes.
          if queue_config.get("pushback_seconds"):
            pushback_time = current_time + float(queue_config.get("pushback_seconds"))
            params = _redis_command_zaddbyscore()(keys=[queue.redis_key], args=["-inf", current_time, 0, max_jobs, pushback_time])
          else:
            params = _redis_command_zpopbyscore()(keys=[queue.redis_key], args=["-inf", current_time, 0, max_jobs])

        # ZSET
        elif queue.is_sorted:

          with connections.redis.pipeline(transaction=True) as pipe:
            pipe.zrange(queue.redis_key, 0, max_jobs - 1)
            pipe.zremrangebyrank(queue.redis_key, 0, max_jobs - 1)
            params = pipe.execute()[0]

        # SET
        elif queue.is_set:
          params = _redis_group_command("spop", max_jobs, queue.redis_key)

        # LIST
        else:
          params = _redis_group_command("lpop", max_jobs, queue.redis_key)

        if len(params) == 0:
          continue

        max_jobs -= len(params)

        if queue.is_raw:
          job_data = [job_factory(p) for p in params]
          for j in job_data:
            j["status"] = "started"
            j["queue"] = queue.id
            if worker:
              j["worker"] = worker.id

          jobs += job_class.insert(job_data)

        else:
          jobs += [job_class(_job_id, queue=queue, start=True)
                   for _job_id in params if _job_id]

        if max_jobs == 0:
          break

      return jobs


def _redis_group_command(command, cnt, redis_key):
  with connections.redis.pipeline(transaction=False) as pipe:
    for _ in range(cnt):
      getattr(pipe, command)(redis_key)
    return [x for x in pipe.execute() if x]


def send_raw_tasks(queue, params_list, **kwargs):
  q = Queue(queue)
  q.enqueue_raw_jobs(params_list, **kwargs)


def send_task(path, params, **kwargs):
  return send_tasks(path, [params], **kwargs)[0]


def send_tasks(path, params_list, queue=None, sync=False, batch_size=1000):

  if len(params_list) == 0:
    return []

  if sync:
    task_class = load_class_by_path(path)
    return [task_class().run(params) for params in params_list]

  if queue is None:
    task_def = get_current_config().get("tasks", {}).get(path) or {}
    queue = task_def.get("queue", "default")

  all_ids = []

  collection = connections.mongodb_jobs.mrq_jobs

  for params_group in group_iter(params_list, n=batch_size):

    job_ids = collection.insert([{
      "path": path,
      "params": params,
      "queue": queue,
      "status": "queued"
    } for params in params_group], w=1)

    # Between these 2 calls, a task can be inserted in MongoDB but not queued in Redis.
    # This is the same as dequeueing a task from Redis and being stopped before updating the "started"
    # flag in MongoDB.

    Queue(queue).enqueue_job_ids([str(x) for x in job_ids])

    all_ids += job_ids

  return all_ids



import argparse
import os
import sys
from .version import VERSION
from .utils import get_local_ip
import atexit


def add_parser_args(parser, config_type):

  # General arguments

  parser.add_argument('--trace_greenlets', action='store_true', default=False,
                      help='Collect stats about each greenlet execution time and switches.')

  parser.add_argument('--trace_memory', action='store_true', default=False,
                      help='Collect stats about memory for each task. Incompatible with gevent > 1')

  parser.add_argument('--trace_mongodb', action='store_true', default=False,
                      help='Collect stats about MongoDB requests')

  parser.add_argument('--print_mongodb', action='store_true', default=False,
                      help='Print all MongoDB requests')

  parser.add_argument('--trace_memory_type', action='store', default="",
                      help='Create a .png object graph in trace_memory_output_dir with a random object of this type.')

  parser.add_argument('--trace_memory_output_dir', action='store', default="memory_traces",
                      help='Directory where to output .pngs with object graphs')

  parser.add_argument('--profile', action='store_true', default=False,
                      help='Run profiling on the whole worker')

  parser.add_argument('--mongodb_jobs', action='store', default="mongodb://127.0.0.1:27017/mrq",
                      help='MongoDB URI for the jobs, scheduled_jobs & workers database')

  parser.add_argument('--mongodb_logs', action='store', default="mongodb://127.0.0.1:27017/mrq",
                      help='MongoDB URI for the logs database. If set to "0", will disable remote logs.')

  parser.add_argument('--mongodb_logs_size', action='store', default=16 * 1024 * 1024, type=int,
                      help='If provided, sets the log collection to capped to that amount of bytes.')

  parser.add_argument('--redis', action='store', default="redis://127.0.0.1:6379",
                      help='Redis URI')

  parser.add_argument('--redis_prefix', action='store', default="mrq",
                      help='Redis key prefix')

  parser.add_argument('--name', default=None, action='store',
                      help='Specify a different name')

  parser.add_argument('--quiet', default=False, action='store_true',
                      help='Don\'t output task logs')

  parser.add_argument('--config', '-c', default=None, action="store",
                      help='Path of a config file')

  parser.add_argument('--worker_class', default="mrq.worker.Worker", action="store",
                      help='Path to a custom worker class')

  parser.add_argument('--version', '-v', default=False, action="store_true",
                      help='Prints current MRQ version')

  parser.add_argument('--no_import_patch', default=False, action='store_true',
                      help='Skips patching __import__ to fix gevent bug #108')

  # mrq-run-specific arguments

  if config_type == "run":

    parser.add_argument('--async', action='store_true', default=False,
                        help='Queue the task instead of running it right away')

    parser.add_argument('--queue', action='store', default="default",
                        help='Queue where to put the task when async')

    parser.add_argument('taskpath', action='store',
                        help='Task to run')

    parser.add_argument('taskargs', action='store', default='{}', nargs='*',
                        help='JSON-encoded arguments, or "key value" pairs')

  # Dashboard-specific arguments

  elif config_type == "dashboard":

    parser.add_argument('--dashboard_httpauth', default="", action="store",
                        help='HTTP Auth for the Dashboard. Format is user:pass')

    parser.add_argument('--dashboard_queue', default=None, action="store",
                        help='Default queue for dashboard actions.')

  # Worker-specific args

  elif config_type == "worker":

    parser.add_argument('--max_jobs', default=0, type=int, action='store',
                        help='Gevent: max number of jobs to do before quitting. Temp workaround for memory leaks')

    parser.add_argument('--gevent', '-g', default=1, type=int, action='store',
                        help='Gevent: max number of greenlets')

    parser.add_argument('--processes', '-p', default=0, type=int, action='store',
                        help='Number of processes to launch with supervisord')

    default_template = os.path.abspath(os.path.join(os.path.dirname(__file__), "supervisord_templates/default.conf"))
    parser.add_argument('--supervisord_template', default=default_template, action='store',
                        help='Path of supervisord template to use')

    parser.add_argument('--scheduler', default=False, action='store_true',
                        help='Run the scheduler')

    parser.add_argument('--scheduler_interval', default=60, action='store', type=int,
                        help='Seconds between scheduler checks')

    parser.add_argument('--report_interval', default=10, action='store', type=int,
                        help='Seconds between worker reports to MongoDB')

    parser.add_argument('queues', nargs='*', default=["default"],
                        help='The queues to listen on (default: \'default\')')

    parser.add_argument('--admin_port', default=0, action="store", type=int,
                        help='Start an admin server on this port, if provided. Incompatible with --processes')

    parser.add_argument('--local_ip', default=get_local_ip(), action="store", type=str,
                        help='Overwrite the local IP, to be displayed in the dashboard.')


def get_config(sources=("file", "env", "args"), env_prefix="MRQ_", parser=None, config_type="worker"):

  if not parser:
    parser = argparse.ArgumentParser()

  add_parser_args(parser, config_type)

  if config_type in ["run"]:
    default_config = parser.parse_args(["x"]).__dict__
  else:
    default_config = parser.parse_args([]).__dict__

  # Keys that can't be passed from the command line
  default_config["tasks"] = {}
  default_config["scheduled_tasks"] = {}

  # Only keep values different from config, actually passed on the command line
  from_args = {}
  if "args" in sources:
    for k, v in parser.parse_args().__dict__.iteritems():
      if default_config[k] != v:
        from_args[k] = v

  # If we were given another config file, use it
  if from_args.get("config"):
    config_file = from_args.get("config")
  # If a mrq-config.py file is in the current directory, use it!
  elif os.path.isfile(os.path.join(os.getcwd(), "mrq-config.py")):
    config_file = os.path.join(os.getcwd(), "mrq-config.py")
  else:
    config_file = None

  from_file = {}
  if config_file and "file" in sources:
    sys.path.insert(0, os.path.dirname(config_file))
    config_module = __import__(os.path.basename(config_file.replace(".py", "")))
    sys.path.pop(0)
    for k, v in config_module.__dict__.iteritems():

      # We only keep variables starting with an uppercase character.
      if k[0].isupper():
        default_config[k.lower()] = v
        if k.lower() not in default_config:
          default_config[k.lower()] = v

  # Merge the config in the order given by the user
  merged_config = default_config
  for part in sources:
    for name in merged_config:

      if part == "env":
        value = os.environ.get(env_prefix + name.upper())
        if value:
          merged_config[name] = value
      elif part == "args" and name in from_args:
        merged_config[name] = from_args[name]
      elif part == "file" and name in from_file:
        merged_config[name] = from_file[name]

  if merged_config["profile"]:
    import cProfile
    profiler = cProfile.Profile()
    profiler.enable()

    def print_profiling():
      profiler.print_stats(sort="cumulative")

    atexit.register(print_profiling)

  if merged_config["version"]:
    print "MRQ version: %s" % VERSION
    print "Python version: %s" % sys.version
    sys.exit(1)

  return merged_config



import json
import datetime
from bson.objectid import ObjectId
# from werkzeug import Response
from functools import wraps
from flask import request, Response
from mrq.context import get_current_config


class MongoJsonEncoder(json.JSONEncoder):
  def default(self, obj):  # pylint: disable-msg=E0202
    if isinstance(obj, (datetime.datetime, datetime.date)):
      return obj.isoformat()
    elif isinstance(obj, ObjectId):
      return unicode(obj)
    return json.JSONEncoder.default(self, obj)


def jsonify(*args, **kwargs):
  """ jsonify with support for MongoDB ObjectId
  """
  return Response(json.dumps(dict(*args, **kwargs), cls=MongoJsonEncoder), mimetype='application/json')


def check_auth(username, pwd):
  """This function is called to check if a username /
  password combination is valid.
  """
  cfg = get_current_config()
  return username == cfg["dashboard_httpauth"].split(":")[0] and pwd == cfg["dashboard_httpauth"].split(":")[1]


def authenticate():
  """Sends a 401 response that enables basic auth"""
  return Response(
    'Could not verify your access level for that URL.\n'
    'You have to login with proper credentials', 401,
    {'WWW-Authenticate': 'Basic realm="Login Required"'}
  )


def requires_auth(f):

  cfg = get_current_config()
  if not cfg["dashboard_httpauth"]:
    return f

  @wraps(f)
  def decorated(*args, **kwargs):
    auth = request.authorization
    if not auth or not check_auth(auth.username, auth.password):
      return authenticate()
    return f(*args, **kwargs)
  return decorated



import datetime
from bson import ObjectId
import pymongo
import time
from .exceptions import RetryInterrupt, CancelInterrupt
from .utils import load_class_by_path
from .queue import Queue
from .context import get_current_worker, log, connections, get_current_config, set_current_job
import gevent
import objgraph
import random
import gc
from collections import defaultdict
import traceback
import sys
from itertools import count as itertools_count


class Job(object):

  # Seconds the job can last before timeouting
  timeout = 300

  # Seconds the results are kept in MongoDB
  result_ttl = 7 * 24 * 3600

  progress = 0

  # Exceptions that don't mark the task as failed but as retry
  retry_on_exceptions = (
    pymongo.errors.AutoReconnect,
    pymongo.errors.OperationFailure,
    pymongo.errors.ConnectionFailure,
    RetryInterrupt
  )

  # Exceptions that will make the task as cancelled
  cancel_on_exceptions = (
    CancelInterrupt
  )

  _memory_stop = 0
  _memory_start = 0

  def __init__(self, job_id, queue=None, start=False, fetch=False):
    self.worker = get_current_worker()
    self.queue = queue
    self.datestarted = None

    self.collection = connections.mongodb_jobs.mrq_jobs

    if job_id is None:
      self.id = None
    else:
      self.id = ObjectId(job_id)

    self.data = None
    self.saved = True

    self.task = None
    self.greenlet_switches = 0
    self.greenlet_time = 0

    self._trace_mongodb = defaultdict(int)

    if start:
      self.fetch(start=True, full_data=False)
    elif fetch:
      self.fetch(start=False, full_data=False)

  def exists(self):
    return bool(self.collection.find_one({"_id": self.id}, fields={"_id": 1}))

  def fetch(self, start=False, full_data=True):
    """ Get the current job data and possibly flag it as started. """

    if self.id is None:
      return self

    if full_data is True:
      fields = None
    elif type(full_data) == dict:
      fields = full_data
    else:
      fields = {
        "_id": 0,
        "path": 1,
        "params": 1,
        "status": 1
      }

    if start:
      self.datestarted = datetime.datetime.utcnow()
      self.data = self.collection.find_and_modify({
        "_id": self.id,
        "status": {"$nin": ["cancel"]}
      }, {"$set": {
        "status": "started",
        "datestarted": self.datestarted,
        "worker": self.worker.id
      }}, fields=fields)
    else:
      self.data = self.collection.find_one({
        "_id": self.id
      }, fields=fields)

    if self.data is None:
      log.info("Job %s not found in MongoDB or status was cancelled!" % self.id)
    else:
      self.populate_properties()

    return self

  def set_progress(self, ratio, save=False):
    self.data["progress"] = ratio
    self.saved = False

    # If not saved, will be updated in the next worker report
    if save:
      self.save()

  def save(self):
    """ Will be called at each worker report. """

    if not self.saved and self.data and "progress" in self.data:
      # TODO should we save more fields?
      self.collection.update({"_id": self.id}, {"$set": {
        "progress": self.data["progress"]
      }})
      self.saved = True

  def populate_properties(self):

    task_def = get_current_config().get("tasks", {}).get(self.data["path"]) or {}
    self.timeout = task_def.get("timeout", self.timeout)
    self.result_ttl = task_def.get("result_ttl", self.result_ttl)

  @classmethod
  def insert(self, jobs_data, queue=None):

    for data in jobs_data:
      if data["status"] == "started":
        data["datestarted"] = datetime.datetime.utcnow()
    connections.mongodb_jobs.mrq_jobs.insert(jobs_data, manipulate=True)

    jobs = []
    for data in jobs_data:
      job = self(data["_id"], queue=queue)
      job.data = data
      job.populate_properties()
      if data["status"] == "started":
        job.datestarted = data["datestarted"]
      jobs.append(job)

    return jobs

  def subpool_map(self, pool_size, func, iterable):
    """ Starts a Gevent pool and run a map. Takes care of setting current_job and cleaning up. """

    if not pool_size:
      return [func(*args) for args in iterable]

    counter = itertools_count()

    def inner_func(*args):
      next(counter)
      set_current_job(self)
      ret = func(*args)
      set_current_job(None)
      return ret

    start_time = time.time()
    pool = gevent.pool.Pool(size=pool_size)
    ret = pool.map(inner_func, iterable)
    pool.join(raise_error=True)
    total_time = time.time() - start_time

    log.debug("SubPool ran %s greenlets in %0.6fs" % (counter, total_time))

    return ret

  def save_status(self, status, result=None, exception=False, dateretry=None, queue=None, w=1):

    if self.id is None:
      return

    now = datetime.datetime.utcnow()
    updates = {
      "status": status,
      "dateupdated": now
    }

    if self.datestarted:
      updates["totaltime"] = (now - self.datestarted).total_seconds()
    if result is not None:
      updates["result"] = result

    if dateretry is not None:
      updates["dateretry"] = dateretry
    if queue is not None:
      updates["queue"] = queue
    if get_current_config().get("trace_greenlets"):
      current_greenlet = gevent.getcurrent()
      updates["time"] = current_greenlet._trace_time
      updates["switches"] = current_greenlet._trace_switches

    if exception:
      trace = traceback.format_exc()
      log.error(trace)
      updates["traceback"] = trace
      updates["exceptiontype"] = sys.exc_info()[0].__name__

    # Make the job document expire
    if status in ("success", "cancel"):
      updates["dateexpires"] = now + datetime.timedelta(seconds=self.result_ttl)

    if status == "success" and "progress" in self.data:
      updates["progress"] = 1

    self.collection.update({
      "_id": self.id
    }, {"$set": updates}, w=w, manipulate=False)

    if self.data:
      self.data.update(updates)

  def save_retry(self, exc, exception=False):

    countdown = 24 * 3600

    if isinstance(exc, RetryInterrupt) and exc.countdown is not None:
      countdown = exc.countdown

    queue = None
    if isinstance(exc, RetryInterrupt) and exc.queue:
      queue = exc.queue

    if countdown == 0:
      self.requeue(queue=queue)
    else:
      self.save_status(
        "retry",
        exception=exception,
        dateretry=datetime.datetime.utcnow() + datetime.timedelta(seconds=countdown),
        queue=queue
      )

  def retry(self, queue=None, countdown=None, max_retries=None):

    if self.task.cancel_on_retry:
      raise CancelInterrupt()
    else:
      exc = RetryInterrupt()
      exc.queue = queue
      exc.countdown = countdown
      raise exc

  def cancel(self):
    self.save_status("cancel")

  def requeue(self, queue=None):

    if not queue:
      if not self.data or not self.data.get("queue"):
        self.fetch(full_data={"_id": 0, "queue": 1, "path": 1})
      queue = self.data["queue"]

    self.save_status("queued", queue=queue)

    # Between these two lines, jobs can become "lost" too.

    Queue(queue).enqueue_job_ids([str(self.id)])

  def perform(self):
    """ Loads and starts the main task for this job, the saves the result. """

    if self.data is None:
      return

    log.debug("Starting %s(%s)" % (self.data["path"], self.data["params"]))
    task_class = load_class_by_path(self.data["path"])

    self.task = task_class()

    self.task.is_main_task = True

    result = self.task.run(self.data["params"])

    self.save_status("success", result)

    if get_current_config().get("trace_greenlets"):

      # TODO: this is not the exact greenlet_time measurement because it doesn't
      # take into account the last switch's time. This is why we force a last switch.
      # This does cause a performance overhead. Instead, we should print the
      # last timing directly from the trace() function in context?

      gevent.sleep(0)
      current_greenlet = gevent.getcurrent()
      log.debug("Job %s success: %0.6fs total, %0.6fs in greenlet, %s switches" % (
        self.id, (datetime.datetime.utcnow() - self.datestarted).total_seconds(),
        current_greenlet._trace_time, current_greenlet._trace_switches - 1)
      )
    else:
      log.debug("Job %s success: %0.6fs total" % (
        self.id, (datetime.datetime.utcnow() - self.datestarted).total_seconds()
      ))

    return result

  def wait(self, poll_interval=1, timeout=None, full_data=False):
    """ Wait for this job to finish. """

    collection = connections.mongodb_jobs.mrq_jobs

    end_time = None
    if timeout:
      end_time = time.time() + timeout

    while (end_time is None or time.time() < end_time):

      job_data = collection.find_one({
        "_id": ObjectId(self.id),
        "status": {"$nin": ["started", "queued"]}
      }, fields=({
        "_id": 0,
        "result": 1,
        "status": 1
      } if not full_data else None))

      if job_data:
        return job_data

      time.sleep(poll_interval)

    raise Exception("Waited for job result for %ss seconds, timeout." % timeout)

  def trace_memory_start(self):
    """ Starts measuring memory consumption """

    objgraph.show_growth(limit=10)

    gc.collect()
    self._memory_start = self.worker.get_memory()

  def trace_memory_stop(self):
    """ Stops measuring memory consumption """

    objgraph.show_growth(limit=10)

    trace_type = get_current_config()["trace_memory_type"]
    if trace_type:

      objgraph.show_chain(
        objgraph.find_backref_chain(
          random.choice(objgraph.by_type(trace_type)),
          objgraph.is_proper_module
        ),
        filename='%s/%s-%s.png' % (get_current_config()["trace_memory_output_dir"], trace_type, self.id)
      )

    gc.collect()
    self._memory_stop = self.worker.get_memory()

    diff = self._memory_stop - self._memory_start

    log.debug("Memory diff for job %s : %s" % (self.id, diff))

    # We need to update it later than the results, we need them off memory already.
    self.collection.update({
      "_id": self.id
    }, {"$set": {
      "memory_diff": diff
    }}, w=1)





from mrq.task import Task
from mrq.context import connections


class MultiRedis(Task):

  def run(self, params):

    connections.redis.set("test", "xxx")
    connections.redis_second.set("test", "yyy")

    assert connections.redis.get("test") == "xxx"
    assert connections.redis_second.get("test") == "yyy"

    return "ok"



from gevent import monkey
monkey.patch_all()

from flask import Flask, request

import time
import os
import sys
from bson import ObjectId
import json
import argparse
from gevent.pywsgi import WSGIServer
from werkzeug.serving import run_with_reloader

sys.path.insert(0, os.getcwd())

from mrq.queue import send_task, Queue
from mrq.context import connections, set_current_config, get_current_config
from mrq.config import get_config

from mrq.dashboard.utils import jsonify, requires_auth

CURRENT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

parser = argparse.ArgumentParser(description='Start the MRQ dashboard')

cfg = get_config(parser=parser, config_type="dashboard")
set_current_config(cfg)

app = Flask("dashboard", static_folder=os.path.join(CURRENT_DIRECTORY, "static"))


@app.route('/')
@requires_auth
def root():
  return app.send_static_file("index.html")


@app.route('/api/datatables/taskexceptions')
@requires_auth
def api_task_exceptions():
  stats = list(connections.mongodb_jobs.mrq_jobs.aggregate([
    {"$match": {"status": "failed"}},
    {"$group": {"_id": {"path": "$path", "exceptiontype": "$exceptiontype"}, "jobs": {"$sum": 1}}},
  ])["result"])

  stats.sort(key=lambda x: -x["jobs"])
  start = int(request.args.get("iDisplayStart", 0))
  end = int(request.args.get("iDisplayLength", 20)) + start

  data = {
    "aaData": stats[start:end],
    "iTotalDisplayRecords": len(stats)
  }

  data["sEcho"] = request.args["sEcho"]

  return jsonify(data)


@app.route('/api/datatables/status')
@requires_auth
def api_jobstatuses():
  stats = list(connections.mongodb_jobs.mrq_jobs.aggregate([
    {"$sort": {"status": 1}},  # https://jira.mongodb.org/browse/SERVER-11447
    {"$group": {"_id": "$status", "jobs": {"$sum": 1}}}
  ])["result"])

  stats.sort(key=lambda x: x["_id"])

  data = {
    "aaData": stats,
    "iTotalDisplayRecords": len(stats)
  }

  data["sEcho"] = request.args["sEcho"]

  return jsonify(data)


@app.route('/api/datatables/taskpaths')
@requires_auth
def api_taskpaths():
  stats = list(connections.mongodb_jobs.mrq_jobs.aggregate([
    {"$sort": {"path": 1}},  # https://jira.mongodb.org/browse/SERVER-11447
    {"$group": {"_id": "$path", "jobs": {"$sum": 1}}}
  ])["result"])

  stats.sort(key=lambda x: -x["jobs"])

  data = {
    "aaData": stats,
    "iTotalDisplayRecords": len(stats)
  }

  data["sEcho"] = request.args["sEcho"]

  return jsonify(data)


@app.route('/workers')
@requires_auth
def get_workers():
  collection = connections.mongodb_jobs.mrq_workers
  cursor = collection.find({"status": {"$ne": "stop"}})
  data = {"workers": list(cursor)}
  return jsonify(data)


def build_api_datatables_query(req):
  query = {}

  if req.args.get("redisqueue"):
    query["_id"] = {"$in": [ObjectId(x) for x in Queue(req.args.get("redisqueue")).list_job_ids(limit=1000)]}
  else:

    for param in ["queue", "path", "status", "exceptiontype"]:
      if req.args.get(param):
        query[param] = req.args.get(param)
    if req.args.get("id"):
      query["_id"] = ObjectId(req.args.get("id"))
    if req.args.get("worker"):
      query["worker"] = ObjectId(req.args.get("worker"))

    if req.args.get("params"):
      try:
        params_dict = json.loads(req.args.get("params"))

        for key in params_dict.keys():
          query["params.%s" % key] = params_dict[key]
      except Exception as e:
        print "Error will converting form JSON: %s" % e

  return query


@app.route('/api/datatables/<unit>')
@requires_auth
def api_datatables(unit):

  # import time
  # time.sleep(5)

  collection = None
  sort = None

  if unit == "queues":

    queues = []
    for name, jobs in Queue.all().items():
      queue = Queue(name)
      q = {
        "name": name,
        "jobs": jobs,  # MongoDB size
        "size": queue.size(),  # Redis size
        "is_sorted": queue.is_sorted,
        "is_timed": queue.is_timed,
        "is_raw": queue.is_raw,
        "is_set": queue.is_set
      }

      if queue.is_sorted:
        q["graph_config"] = cfg.get("raw_queues", {}).get(name, {}).get("dashboard_graph", lambda: {
          "start": time.time() - (7 * 24 * 3600),
          "stop": time.time() + (7 * 24 * 3600),
          "slices": 30
        } if queue.is_timed else {
          "start": 0,
          "stop": 100,
          "slices": 30
        })()
        if q["graph_config"]:
          q["graph"] = queue.get_sorted_graph(**q["graph_config"])

      if queue.is_timed:
        q["jobs_to_dequeue"] = queue.count_jobs_to_dequeue()

      queues.append(q)

    queues.sort(key=lambda x: -(x["jobs"] + x["size"]))

    data = {
      "aaData": queues,
      "iTotalDisplayRecords": len(queues)
    }

  if unit == "workers":
    fields = None
    query = {"status": {"$nin": ["stop"]}}
    collection = connections.mongodb_jobs.mrq_workers
    sort = [("datestarted", -1)]

    if request.args.get("showstopped"):
      query = {}

  elif unit == "scheduled_jobs":
    collection = connections.mongodb_jobs.mrq_scheduled_jobs
    fields = None
    query = {}

  elif unit == "jobs":

    fields = None
    query = build_api_datatables_query(request)
    sort = [("_id", 1)]

    # We can't search easily params because we store it as decoded JSON in mongo :(
    # Add a string index?
    # if request.args.get("sSearch"):
    #   query.update(json.loads(request.args.get("sSearch")))
    collection = connections.mongodb_jobs.mrq_jobs

  if collection is not None:

    cursor = collection.find(query, fields=fields)

    if sort:
      cursor.sort(sort)

    data = {
      "aaData": list(cursor.skip(int(request.args.get("iDisplayStart", 0))).limit(int(request.args.get("iDisplayLength", 20)))),
      "iTotalDisplayRecords": collection.find(query).count()
    }

  data["sEcho"] = request.args["sEcho"]

  return jsonify(data)


@app.route('/api/job/<job_id>/result')
@requires_auth
def api_job_result(job_id):
  collection = connections.mongodb_jobs.mrq_jobs

  job_data = collection.find_one({"_id": ObjectId(job_id)}, fields=["result"])
  if not job_data:
    return jsonify({})

  return jsonify({
    "result": job_data.get("result")
  })


@app.route('/api/job/<job_id>/traceback')
@requires_auth
def api_job_traceback(job_id):
  collection = connections.mongodb_jobs.mrq_jobs
  job_data = collection.find_one({"_id": ObjectId(job_id)}, fields=["traceback"])

  if not job_data:
    job_data = {}

  return jsonify({
    "traceback": job_data.get("traceback", "No exception raised")
  })


@app.route('/api/jobaction', methods=["POST"])
@requires_auth
def api_job_action():
  return jsonify({
    "job_id": send_task("mrq.basetasks.utils.JobAction", {k: v for k, v in request.form.iteritems()}, queue=get_current_config()["dashboard_queue"])
  })


@app.route('/api/logs')
@requires_auth
def api_logs():
  collection = connections.mongodb_logs.mrq_logs

  if request.args.get("job"):
    query = {"job": ObjectId(request.args.get("job"))}
  elif request.args.get("worker"):
    query = {"worker": ObjectId(request.args.get("worker"))}
  else:
    raise Exception("No ID")

  if request.args.get("last_log_id"):
    query["_id"] = {"$gt": ObjectId(request.args.get("min_log_id"))}

  logs = list(collection.find(query, fields={"_id": 1, "logs": 1}))

  data = {
    "logs": "\n".join([lines["logs"] for lines in logs]),

    # Don't be surprised, this will send unexisting ObjectIds when we're up to date!
    "last_log_id": logs[-1]["_id"] if len(logs) else ObjectId()
  }

  return jsonify(data)


def main():
  app.debug = True
  http = WSGIServer(('', int(os.environ.get("PORT", 5555))), app)
  run_with_reloader(http.serve_forever)


if __name__ == '__main__':
  main()



from mrq.task import Task
from mrq.job import Job
from mrq.queue import Queue
from bson import ObjectId
from mrq.context import connections, get_current_config
from collections import defaultdict
from mrq.utils import group_iter
import datetime
import ujson as json


class JobAction(Task):

  def run(self, params):

    self.params = params
    self.collection = connections.mongodb_jobs.mrq_jobs

    query = self.build_query()

    return self.perform_action(self.params.get("action"), query)

  def build_query(self):
    query = {}
    if self.params.get("id"):
      query["_id"] = ObjectId(self.params.get("id"))

    for k in ["queue", "status", "worker", "path", "dateretry", "exceptiontype"]:  # TODO use redis for queue
      if self.params.get(k):
        query[k] = self.params.get(k)

    if self.params.get("params"):
      params_dict = json.loads(self.params.get("params"))

      for key in params_dict.keys():
        query["params.%s" % key] = params_dict[key]

    return query

  def perform_action(self, action, query):

    stats = {
      "requeued": 0,
      "cancelled": 0
    }

    if action == "cancel":

      # Finding the ttl here to expire is a bit hard because we may have mixed paths
      # and hence mixed ttls.
      # If we are cancelling by path, get this ttl
      if query.get("path"):

        task_def = get_current_config().get("tasks", {}).get(query["path"]) or {}
        result_ttl = task_def.get("result_ttl", Job.result_ttl)

      # If not, get the maxmimum ttl of all tasks.
      else:
        result_ttl = max([Job.result_ttl] + [task_def.get("result_ttl", Job.result_ttl) for task_def in get_current_config().get("tasks", {}).values()])

      now = datetime.datetime.utcnow()
      ret = self.collection.update(query, {"$set": {
        "status": "cancel",
        "dateexpires": now + datetime.timedelta(seconds=result_ttl),
        "dateupdated": now
      }}, multi=True)
      stats["cancelled"] = ret["n"]

      # Special case when emptying just by queue name: empty it directly!
      # In this case we could also loose some jobs that were queued after
      # the MongoDB update. They will be "lost" and requeued later like the other case
      # after the Redis BLPOP
      if query.keys() == ["queue"]:
        Queue(query["queue"]).empty()

    elif action == "requeue":

      # Requeue task by groups of maximum 1k items (if all in the same queue)
      cursor = self.collection.find(query, fields=["_id", "queue"])

      # We must freeze the list because queries below would change it.
      # This could not fit in memory, research adding {"stats": {"$ne": "queued"}} in the query
      fetched_jobs = list(cursor)

      for jobs in group_iter(fetched_jobs, n=1000):

        jobs_by_queue = defaultdict(list)
        for job in jobs:
          jobs_by_queue[job["queue"]].append(job["_id"])
          stats["requeued"] += 1

        for queue in jobs_by_queue:
          self.collection.update({
            "_id": {"$in": jobs_by_queue[queue]}
          }, {"$set": {
            "status": "queued",
            "dateupdated": datetime.datetime.utcnow()
          }}, multi=True)

          # Between these two lines, jobs can become "lost" too.

          Queue(queue).enqueue_job_ids([str(x) for x in jobs_by_queue[queue]])

    return stats



from .context import get_current_job


def patch_pymongo(config):
  """ Monkey-patch pymongo's collections to add some logging """

  # Nothing to change!
  if not config["print_mongodb"] and not config["trace_mongodb"]:
    return

  # Print because we are very early and log() may not be ready yet.
  print "Monkey-patching MongoDB methods..."

  from termcolor import cprint

  def gen_monkey_patch(base_object, method):
    base_method = getattr(base_object, method)

    def monkey_patched(self, *args, **kwargs):
      if config["print_mongodb"]:
        if self.full_name in config.get("print_mongodb_hidden_collections", []):
          cprint("[MONGO] %s.%s%s %s" % (self.full_name, method, "-hidden-", kwargs), "magenta")
        else:
          cprint("[MONGO] %s.%s%s %s" % (self.full_name, method, args, kwargs), "magenta")

      if config["trace_mongodb"]:
        job = get_current_job()
        if job:
          job._trace_mongodb[method] += 1
      return base_method(self, *args, **kwargs)

    return monkey_patched

  from pymongo.collection import Collection
  for method in ["find", "update", "insert", "remove", "find_and_modify"]:
    setattr(Collection, method, gen_monkey_patch(Collection, method))

  # MongoKit completely replaces the code from PyMongo's find() function, so we
  # need to monkey-patch that as well.
  try:
    from mongokit.collection import Collection as MongoKitCollection
    for method in ["find"]:
      setattr(MongoKitCollection, method, gen_monkey_patch(MongoKitCollection, method))

  except ImportError:
    pass


# https://code.google.com/p/gevent/issues/detail?id=108
def patch_import():

  import types
  import gevent.coros
  import __builtin__

  orig_import = __builtin__.__import__
  import_lock = gevent.coros.RLock()

  def safe_import(*args, **kwargs):
    """
    Normally python protects imports against concurrency by doing some locking
    at the C level (at least, it does that in CPython).  This function just
    wraps the normal __import__ functionality in a recursive lock, ensuring that
    we're protected against greenlet import concurrency as well.
    """
    if len(args) > 0 and type(args[0]) not in [
      types.StringType,
      types.UnicodeType]:
      # if a builtin has been acquired as a bound instance method,
      # python knows not to pass 'self' when the method is called.
      # No such protection exists for monkey-patched builtins,
      # however, so this is necessary.
      args = args[1:]
    import_lock.acquire()
    try:
      result = orig_import(*args, **kwargs)
    finally:
      import_lock.release()
    return result

  builtins = __import__('__builtin__')
  builtins.__import__ = safe_import



from time import sleep
from mrq.task import Task
from mrq.context import log, retry_current_job, connections, get_current_config, get_current_job, progress
import urllib2


class Add(Task):

  def run(self, params):

    log.info("adding", params)
    res = params.get("a", 0) + params.get("b", 0)

    if params.get("sleep", 0):
      log.info("sleeping", params.get("sleep", 0))
      sleep(params.get("sleep", 0))

    return res


class TimeoutFromConfig(Add):
  pass


class TimeoutFromConfigAndCancel(Add):
  cancel_on_timeout = True


class Fetch(Task):
  def run(self, params):

    f = urllib2.urlopen(params.get("url"))
    t = f.read()
    f.close()

    return len(t)


class GetConfig(Task):
  def run(self, params):
    return get_current_config()


LEAKS = []


class Leak(Task):
  def run(self, params):

    if params.get("size", 0) > 0:
      #LEAKS.append("".join([random.choice(string.letters) for _ in range(params.get("size", 0))]))
      LEAKS.append(["1" for _ in range(params.get("size", 0))])

    if params.get("sleep", 0) > 0:
      sleep(params.get("sleep", 0))

    return params.get("return")


class Retry(Task):

  def run(self, params):

    log.info("Retrying in %s on %s" % (params.get("countdown"), params.get("queue")))

    connections.mongodb_logs.tests_inserts.insert(params)

    if params.get("cancel_on_retry"):
      self.cancel_on_retry = params.get("cancel_on_retry")

    retry_current_job(queue=params.get("queue"), countdown=params.get("countdown"))

    raise Exception("Should not be reached")


class RaiseException(Task):

  def run(self, params):

    sleep(params.get("sleep", 0))

    raise Exception(params.get("message", ""))


class ReturnParams(Task):

  def run(self, params):

    sleep(params.get("sleep", 0))

    return params


class Progress(Task):

  def run(self, params):

    for i in range(1, 100):
      progress(0.01 * i, save=params["save"])
      sleep(0.1)


class MongoInsert(Task):

  def run(self, params):

    connections.mongodb_logs.tests_inserts.insert({"params": params}, manipulate=False)

    if params.get("sleep", 0) > 0:
      sleep(params.get("sleep", 0))

    return params


MongoInsert2 = MongoInsert


class SubPool(Task):

  def inner(self, x):

    assert get_current_job() == self.job

    if x == "import-large-file":
      from mrq.basetasks.tests.largefile import a
      assert a == 1
      return True

    if x == "exception":
      raise Exception(x)

    sleep(x)

    return x

  def run(self, params):

    self.job = get_current_job()

    return self.job.subpool_map(params["pool_size"], self.inner, params["inner_params"])



from mrq.task import Task
from mrq.context import log


class Simple(Task):

  def run(self, params):

    # Some systems may be configured like this.
    if params.get("utf8_sys_stdout"):
      import codecs, sys
      UTF8Writer = codecs.getwriter('utf8')
      sys.stdout = UTF8Writer(sys.stdout)

    if params["class_name"] == "unicode":
      log.info(u"caf\xe9")
    elif params["class_name"] == "string":
      log.info("cafe")
    elif params["class_name"] == "latin-1":
      log.info("caf\xe9")
    elif params["class_name"] == "bytes1":
      log.info("Mat\xc3\xa9riels d'entra\xc3\xaenement")

    return True
